# -*- coding: utf-8 -*-

###########################################################
# PyToxo
#
# A Python tool to calculate penetrance tables for 
# high-order epistasis models
#
# Copyright 2021 Borja González Seoane
#
# Contact: borja.gseoane@udc.es
###########################################################

"""Epistasis model definition."""

import itertools
import os
import string
from typing import Dict, List, Tuple, Union

import mpmath
import numpy
import sympy

import pytoxo.calculations
import pytoxo.errors
import pytoxo.ptable
import pytoxo.util

_MPMATH_DEFAULT_DPS = 15
_TOLERABLE_SOLUTION_ERROR_BASE_DELTA = 1e-16  # It is fitted then to model's order
_TOLERABLE_SOLUTION_ERROR_MAX_DELTA = 1e-8


class Model:
    """Representation of an epistasis model."""

    def __init__(
        self,
        filename: str = None,
        genotypes_dict: Dict[str, str] = None,
        definitions: Union[List[str], numpy.array] = None,
        probabilities: Union[List[str], numpy.array] = None,
        model_name: str = None,
    ):
        """Initializes a model object.

        This constructor can be used in three modes:

        1. Generating the model object from its representation as a CSV file.
        2. Generating the model object from a dictionary with the genotype
            definitions and the expressions of the probabilities of the
            genotypes (`genotypes_dict`).
        3. Generating the model object from two separated lists or Numpy
            arrays with the genotype definitions and the expressions of the
            probabilities of the genotypes (`definitions` and `probabilities`).

        Generating from its representation as a CSV file
        ------------------------------------------------

        Uses the model from its text representation in `filename` file and
        inits an object with its data.

        The input model must be formatted as a plain CSV, with each line of the
        file corresponding to a row of the model. The rows are made of the
        genotype definition and the probability associated with the given
        genotype, separated by a comma.

        Probability is expressed as a function of two variables, represented
        as alphabetical characters. Only two variables are supported at most.
        Any pair of alphabetic characters can be used to represent these
        variables, but they must be the same two names for the entire model
        file. Math syntax should be the used by Sympy.

        Empty lines, as well as lines starting with '#' will be ignored.

        To use this constructor method, use the `filename` parameter
        and let others as default `None` but optional `model_name`.


        Generating from a dictionary of genotype definitions and probabilities
        ----------------------------------------------------------------------

        Uses a Python dictionary, `genotypes_dict`, with the genotypes
        definitions and its associated probabilities, both as strings. E.g.
        of a member of the dict: `"AABBCc": "x*(1+y)"`.

        Probability is expressed as a function of two variables, represented
        as alphabetical characters. Only two variables are supported at most.
        Any pair of alphabetic characters can be used to represent these
        variables, but they must be the same two names for the entire model
        file. Math syntax should be the used by Sympy.

        To use this constructor method, use the `genotypes_dict` parameter
        and let others as default `None` but optional `model_name`.


        Generating from two separated Python lists or Numpy arrays
        ----------------------------------------------------------

        Uses two separated Python lists or Numpy arrays, `definitions` and
        `probabilities`, with the genotypes definitions and its associated
        probabilities, respectively, both as strings. The link between the
        members of both sets is determined by the order. Examples of the two
        sets, respectively: "AABBCc" and "x*(1+y)".

        Probability is expressed as a function of two variables, represented
        as alphabetical characters. Only two variables are supported at most.
        Any pair of alphabetic characters can be used to represent these
        variables, but they must be the same two names for the entire model
        file. Math syntax should be the used by Sympy.

        To use this constructor method, use the `definitions` and
        `probabilities` parameters and let others as default `None` but
        optional `model_name`.


        Parameters
        ----------
        filename : str, optional
            The path of the text file with the model.
        genotypes_dict : dict[str, str], optional
            Dict with genotypes definitions and its associated probabilities.
        definitions: Union[list[str], numpy.array], optional
            List ot Numpy array with the genotype definitions as strings.
            Should be used with `probabilities` parameter.
        probabilities: Union[list[str], numpy.array], optional
            List ot Numpy array with the genotype probabilities as strings.
            Should be used with `definitions` parameter.
        model_name: str, optional
            The name to identify the model. Optional. Using the `filename`
            can be automatically deduced attending to the file name, if this
            parameter is not used. Using the `genotypes_dict` constructor and
            letting this parameter unused the result name would be "unnamed".

        Raises
        ------
        pytoxo.errors.BadFormedModelError
            If the parse tentative fails due to the file is not well formed.
        ValueError
            Bad parameters combination.
        IOError
           If the input file is not found or the reading tentative fails due
           to other unexpected operative system level cause.
        """
        # Parse parameters use to determine constructor way to use. Ignore
        # for now `model_name`
        exclusive_set_params = [
            True
            for p in [filename, genotypes_dict, definitions, probabilities]
            if p is not None
        ]
        if len(exclusive_set_params) == 1 and filename:
            constructor_mode = 1
        elif len(exclusive_set_params) == 1 and genotypes_dict:
            constructor_mode = 2
        elif (
            len(exclusive_set_params) == 2
            and definitions
            is not None  # Need to be explicit because could be Numpy array
            and probabilities
            is not None  # Need to be explicit because could be Numpy array
        ):
            constructor_mode = 3
        else:
            raise ValueError("Bad parameters use. Revise documentation.")

        # Basic skeleton
        self._name = None
        self._order = None  # Number of loci defined in the model
        self._penetrances = (
            []
        )  # List of symbolic expressions representing the epistatic model
        self._variables = (
            []
        )  # List of symbolic variable names used throughout the model

        # Check custom model name
        if model_name:
            self._name = str(model_name)

        # Delegates object composition attending to the used constructor way
        if constructor_mode == 1:
            self._parse_model_file(filename)
        elif constructor_mode == 2:
            self._parse_genotypes_dict(genotypes_dict)
        elif constructor_mode == 3:
            self._parse_genotypes_sets(definitions, probabilities)

    def _parse_model_file(self, filename: str) -> None:
        """Takes the responsibility of the initializer to parse the model file.

        Reads the model from its text representation in `filename` file and
        inits the object with its data.

        The input model must be formatted as a plain CSV, with each line of the
        file corresponding to a row of the model. The rows are made of the
        genotype definition and the probability associated with the given
        genotype, separated by a comma.

        Probability is expressed as a function of two variables, represented
        as alphabetical characters. Only two variables are supported at most.
        Any pair of alphabetic characters can be used to represent these
        variables, but they must be the same two names for the entire model
        file. Math syntax should be the used by Sympy.

        Empty lines, as well as lines starting with '#' will be ignored.

        Parameters
        ----------
        filename : str
            The path of the text file with the model.

        Raises
        ------
        pytoxo.errors.BadFormedModelError
            If the parse tentative fails due to the file is not well formed.
        IOError
           If the input file is not found or the reading tentative fails due
           to other unexpected operative system level cause.
        """
        try:
            with open(filename, "r") as f:
                lines = f.readlines()

            # Discard comments and empty lines
            lines = [line for line in lines if not line.startswith("#")]
            lines = [line for line in lines if line.strip()]

            # Check not empty file
            if not lines:
                raise pytoxo.errors.BadFormedModelError(
                    filename, "File without content."
                )

            # Split lines around the comma (',')
            fst_members = [line.split(",")[0].strip() for line in lines]
            snd_members = [line.split(",")[1].strip() for line in lines]

            # Save the name of the model, if a custom one is not used
            if not self._name:
                self._name = os.path.basename(filename).split(".")[0]

            # Delegate to the helper
            self._parse_helper(
                genotypes=fst_members,
                probabilities=snd_members,
                exception_object_to_raise=filename,
            )

        except IOError as e:
            raise e
        except pytoxo.errors.BadFormedModelError as e:
            raise e
        except:  # Generic drain for unchecked parsing errors
            raise pytoxo.errors.BadFormedModelError(filename)

    def _parse_genotypes_dict(self, genotypes_dict: Dict[str, str]) -> None:
        """Takes the responsibility of the initializer to parse the
        genotypes dict.

        Uses a Python dictionary, `genotypes_dict`, with the genotypes
        definitions and its associated probabilities, both as strings. E.g.
        of a member of the dict: `"AABBCc": "x*(1+y)"`.

        Probability is expressed as a function of two variables, represented
        as alphabetical characters. Only two variables are supported at most.
        Any pair of alphabetic characters can be used to represent these
        variables, but they must be the same two names for the entire model
        file. Math syntax should be the used by Sympy.

        Parameters
        ----------
        genotypes_dict : dict[str, str]
            Dict with genotypes definitions and its associated probabilities.

        Raises
        ------
        pytoxo.errors.BadFormedModelError
            If the input genotypes dict is bad-formed.
        """
        try:
            # Delegate to the helper
            self._parse_helper(
                genotypes=list(genotypes_dict.keys()),
                probabilities=list(genotypes_dict.values()),
                exception_object_to_raise=genotypes_dict,
            )

            # Revise unnamed models
            if not self._name:
                self._name = "unnamed"

        except pytoxo.errors.BadFormedModelError as e:
            raise e
        except:  # Generic drain for unchecked parsing errors
            raise pytoxo.errors.BadFormedModelError(genotypes_dict)

    def _parse_genotypes_sets(
        self,
        definitions: Union[List[str], numpy.array],
        probabilities: Union[List[str], numpy.array],
    ) -> None:
        """Takes the responsibility of the initializer to parse the
        two sets with genotype definitions and probabilities.

        Uses two separated Python lists or Numpy arrays, `definitions` and
        `probabilities`, with the genotypes definitions and its associated
        probabilities, respectively, both as strings. The link between the
        members of both sets is determined by the order. Examples of the two
        sets, respectively: "AABBCc" and "x*(1+y)".

        Probability is expressed as a function of two variables, represented
        as alphabetical characters. Only two variables are supported at most.
        Any pair of alphabetic characters can be used to represent these
        variables, but they must be the same two names for the entire model
        file. Math syntax should be the used by Sympy.

        To use this constructor method, use the `definitions` and
        `probabilities` parameters and let others as default `None` but
        optional `model_name`.


        Parameters
        ----------
        definitions: Union[list[str], numpy.array]
            List ot Numpy array with the genotype definitions as strings.
            Should be used with `probabilities` parameter.
        probabilities: Union[list[str], numpy.array]
            List ot Numpy array with the genotype probabilities as strings.
            Should be used with `definitions` parameter.

        Raises
        ------
        pytoxo.errors.BadFormedModelError
            If the input genotype sets are bad-formed.
        """
        try:
            # Transform Numpy arrays to lists
            if type(definitions) == numpy.array:
                definitions = definitions.tolist()
            if type(probabilities) == numpy.array:
                probabilities = probabilities.tolist()

            # Delegate to the helper
            self._parse_helper(
                genotypes=list(definitions),
                probabilities=list(probabilities),
                exception_object_to_raise=(definitions, probabilities),
            )

            # Revise unnamed models
            if not self._name:
                self._name = "unnamed"

        except pytoxo.errors.BadFormedModelError as e:
            raise e
        except:  # Generic drain for unchecked parsing errors
            raise pytoxo.errors.BadFormedModelError((definitions, probabilities))

    def _parse_helper(
        self,
        genotypes: List[str],
        probabilities: List[str],
        exception_object_to_raise: Union[str, Dict[str, str], Tuple[list, list]],
    ) -> None:
        """Complete the parse process doing the needed checks and
        transformations.

        Parameters
        ----------
        genotypes: list[str]
            List with the genotype definitions as strings.
        probabilities: list[str]
            List with the genotype associated probability expressions as strings.
        exception_object_to_raise: Union[str, dict[str, str], tuple[list, list]]
            Object to add to the exception body.

        Raises
        ------
        pytoxo.errors.BadFormedModelError
            On any fail.
        """
        # Catch the order of the model
        len_genotype = len(genotypes[0])
        # Assert all the first members have the same length
        for genotype in genotypes:
            if len(genotype) != len_genotype:
                raise pytoxo.errors.BadFormedModelError(
                    exception_object_to_raise,
                    "All the genotypes should have the same order.",
                )
        # Assert first members length is odd
        if len_genotype % 2 != 0:
            raise pytoxo.errors.BadFormedModelError(
                exception_object_to_raise, "Bad genotype specification."
            )
        # Save the order
        self._order = len_genotype // 2

        """Sort probabilities attending to alphabetical sort of associated 
        genotypes. Capital letters first. This is necessary to assert the 
        association between genotype definitions and probabilities during the 
        calculus process and in the final penetrance table"""
        genotypes_probabilities = [(g, p) for g, p in zip(genotypes, probabilities)]
        genotypes_probabilities.sort(key=lambda i: i[0])  # Sort attending to genotypes
        probabilities_sorted = [p for (_, p) in genotypes_probabilities]
        # Save the penetrances as symbolic expressions
        try:
            self._penetrances = [
                sympy.sympify(probability) for probability in probabilities_sorted
            ]
        except sympy.SympifyError:
            raise pytoxo.errors.BadFormedModelError(
                exception_object_to_raise, "Bad probability expression syntax."
            )

        # Save the variables of the used expressions
        all_variables = []
        for probability in probabilities_sorted:
            all_variables.append(sympy.symbols([i for i in probability if i.isalpha()]))

        # Reduce the variables to avoid replication
        all_variables = [item for subl in all_variables for item in subl]  # Flat list
        different_variables = []
        for variable in all_variables:
            """Not use other more direct method like
            `list(set(all_variables))` because the variable apparition
            order must be preserved"""
            if variable not in different_variables:
                different_variables.append(variable)

        # Check support: only 2 different variables are supported by PyToxo
        if not 0 < len(different_variables) <= 2:
            raise pytoxo.errors.BadFormedModelError(
                exception_object_to_raise,
                "Only two variables are supported at most.",
            )
        else:
            self._variables = different_variables

        # Calculate tolerable solution error delta
        self._tolerable_solution_error_delta = (
            self.calculate_tolerable_solution_error_delta()
        )

    def calculate_tolerable_solution_error_delta(self) -> float:
        """Calculates the error delta to be tolerated during the model
        resolution, adjusted to the current model order. If the adjusted
        delta is greater than the maximum delta, returns the maximum delta.
        The latter avoid absurd delta error for very large models.

        Uses the constants `_TOLERABLE_SOLUTION_ERROR_BASE_DELTA` and
        `_TOLERABLE_SOLUTION_ERROR_MAX_DELTA` to calculate the fitted delta.

        Returns
        -------
        float
            The resolution error delta to tolerate.
        """
        return min(
            _TOLERABLE_SOLUTION_ERROR_BASE_DELTA * pow(10, self._order),
            _TOLERABLE_SOLUTION_ERROR_MAX_DELTA,
        )

    ########################################
    # Getters and setters for properties

    def _get_name(self) -> str:
        return self._name

    def _set_name(self, name: str) -> None:
        self._name = name

    name = property(_get_name, _set_name)

    @property
    def order(self) -> int:
        return self._order

    @property
    def penetrances(self) -> List[Union[float, sympy.Expr]]:
        return self._penetrances

    @property
    def variables(self) -> List[sympy.Symbol]:
        return self._variables

    @property
    def tolerable_solution_error_delta(self) -> float:
        return self._tolerable_solution_error_delta

    ########################################

    def __hash__(self):
        return hash(
            hash(self._name)
            + hash(self._order)
            + hash(str(self._penetrances))
            + hash(str(self._variables))
            + hash(str(self._tolerable_solution_error_delta))
        )

    def __eq__(self, other):
        return hash(self) == hash(other)

    def calculate_genotypes(self) -> List[str]:
        """Calculate the list of genotypes to the given model attending to
        the model order.

        Uses default alphabetical sort. Capital letters first.

        Returns
        -------
        list[str]
            List of genotypes.
        """
        # Deduce alleles attending to the table order
        # TODO: Is `len(ascii_lowercase) < self._order` possible?
        letters = list(string.ascii_lowercase[: self._order])
        alleles = []
        for letter in letters:
            lower = letter
            upper = letter.upper()
            alleles.append([f"{upper}{upper}", f"{upper}{lower}", f"{lower}{lower}"])
        # Generate genotypes tracing the alleles cartesian product
        return [str("".join(p)) for p in list(itertools.product(*alleles))]

    def _max_penetrance(self) -> sympy.Expr:
        """Returns the largest of all penetrance expressions, for any real
        and positive value of the two variables and attending to the
        mathematical restrictions of the model."""

        # First, remove duplicate expressions to evaluate them
        unique_penetrances = list(set(self._penetrances))

        """The expressions must be monotonically non-decreasing and 
        sortable when `x` and `y` are real positive numbers. So, 
        to achieve the largest polynomial within the penetrance expressions,
        simply does a substitution for `x` and `y` to real positive numbers, and
        the largest numerical reduction will represent also the largest 
        symbolic expression."""
        reductions = [
            p.subs({self._variables[0]: 1, self._variables[1]: 1})
            for p in unique_penetrances
        ]  # 1 is real and positive
        return unique_penetrances[reductions.index((max(reductions)))]

    def _solve(
        self, eq_system: List[sympy.Eq], solve_timeout: Union[int, bool] = True
    ) -> Dict[sympy.Symbol, float]:
        """Assumes the responsibility of solving the system of equations that
        will define the values of the variables for the generation of the
        penetrance table.

        `find_max_prevalence` and `find_max_heritability` delegates here
        after conform their equations systems.

        Parameters
        ----------
        eq_system : list[sympy.Eq]
            Input equations to solve.
        solve_timeout : Union[int, bool], optional (default `True`)
            A maximum timeout, as seconds, for the solver to try to resolve the
            model. If it is exceeded, the operation will be aborted. Default
            (passed as 'True') is assumed since an heuristic method related
            to the model order. Pass as 'False' or 'None' to do not use
            timeout.

        Returns
        -------
        dict[sympy.Symbol: float]
            The equation solution for the two model variables as a dict.

        Raises
        ------
        pytoxo.errors.ResolutionError
            If PyToxo is not able to solve the model.
        pytoxo.errors.UnsolvableModelError
            If the model has not solution.
        """
        # Check timeout
        if not solve_timeout:
            solve_timeout = None  # Required exactly `None`
        elif type(solve_timeout) == bool and solve_timeout == True:
            solve_timeout = (self._order + 1) ** 2 * 60  # Seconds

        # TODO: Consider add assumptions to vars real and greather than 0

        def try_to_solve(relax_dps: int = None) -> Union[Tuple[float], None]:
            """Tries solve the given `eq_system` equations with Sympy and
            returns solutions, filtering unreal and negative ones. The call
            to the solver is protected with a timeout to prevent
            non-termination issues of unsolvable models.

            Parameters
            ----------
            relax_dps : int (default None)
                This parameter contains the decimal digits of precision (DPS) to
                use in the internal operations delegated to MPMath library. By
                default, this library uses 15 digits. To use this value you
                could pass a 15, but better simply leave this value as a `None`
                as default. If this function does not work for you and you
                think that the model is solvable, then you could try 10 or 7,
                which have proven to be values with which the library manages
                to solve certain of the most complex models. Keep in mind that
                if you go overboard with downgrading the precision, the
                calculations could get corrupted.
            """

            @pytoxo.util.timeout(solve_timeout)
            def solver_call():
                """Help function which encapsulate the call to the solver
                with a timeout decorator to abort futile executions of
                unsolvable models."""
                return sympy.solve(
                    eq_system,
                    self._variables[0],
                    self._variables[1],
                    manual=True,
                    rational=False,
                )

            # If applies, relax MPMath DPS to error tolerance
            if relax_dps:
                mpmath.mp.dps = relax_dps

            # Try to solve the system within the setting timeout
            try:
                sols = solver_call()
            except TimeoutError:
                raise pytoxo.errors.ResolutionError(
                    cause="Exceeded timeout",
                    model_name=self.name,
                    equation=str(eq_system),
                )
            finally:
                # Re-adjust MPMath DPS
                if relax_dps:
                    mpmath.mp.dps = _MPMATH_DEFAULT_DPS

            # Discard unreal solutions
            sols = [s for s in sols if s[0].is_real and s[1].is_real]
            # Discard negative solutions
            sols = [s for s in sols if s[0] > 0 and s[1] > 0]
            # Return only one solution
            if sols:
                return sols[0]
            else:
                return None

        try:
            sol = try_to_solve()
            if not sol:
                raise pytoxo.errors.UnsolvableModelError(
                    model_name=self.name, equation=str(eq_system)
                )
        except mpmath.mp.NoConvergence:
            # Try a second time relaxing error tolerance
            try:
                sol = try_to_solve(relax_dps=10)
                if not sol:
                    raise pytoxo.errors.UnsolvableModelError(
                        model_name=self.name, equation=str(eq_system)
                    )
            except mpmath.mp.NoConvergence:
                raise pytoxo.errors.ResolutionError(
                    model_name=self.name, equation=str(eq_system)
                )

        # Errors 'UnsolvableModelError' and 'ResolutionError' are raised directly
        except pytoxo.errors.UnsolvableModelError as e:
            raise e
        except pytoxo.errors.ResolutionError as e:
            raise e

        # Wildcard drain for an unchecked situation
        except:
            raise pytoxo.errors.ResolutionError(
                model_name=self.name, equation=str(eq_system)
            )

        # Return the final achieved solution as peenetrance table object
        return {self._variables[0]: sol[0], self._variables[1]: sol[1]}

    def _check_solution(
        self, eq_system: List[sympy.Eq], sol: Dict[sympy.Symbol, float]
    ) -> Tuple[bool, float]:
        """Check if an achieved solution is correct.

        Parameters
        ----------
        eq_system : list[sympy.Eq]
            Equations previously solved. Right hand sides are the values
            against compare left hand sides substituted.
        sol : dict[sympy.Symbol : float]
            Achieved solution to check.

        Returns
        -------
        tuple[bool, float]
            Verification about the solution as boolean and delta desviation
            if applies.
        """
        eq1 = eq_system[0]
        eq2 = eq_system[1]

        # Substitute the solution in both equations
        eq1_lhs_eval = eq1.lhs.subs(sol).evalf()
        eq2_lhs_eval = eq2.lhs.subs(sol).evalf()

        # Calculate deltas against right hand sides
        delta1 = abs(eq1.rhs.evalf() - eq1_lhs_eval)
        delta2 = abs(eq2.rhs.evalf() - eq2_lhs_eval)

        # Check is deltas respect tolerance
        if (
            delta1 > self._tolerable_solution_error_delta
            or delta2 > self._tolerable_solution_error_delta
        ):
            return False, max(delta1, delta2)
        else:
            return True, 0

    def _build_max_prevalence_system(
        self, mafs: List[float], h: float
    ) -> List[sympy.Eq]:
        """Builds the system of equations for this model, for a maximum
        prevalence, for the given MAFs and heritability.

        Parameters
        ----------
        mafs : list[float]
            Array of floats representing the MAF of each locus.
        h : float
            Heritability of the table.

        Returns
        -------
        list[sympy.Eq]
            System of equations to solve this model for a maximum
            prevalence, for the given MAFs and heritability.

        """
        eq1 = sympy.Eq(
            pytoxo.calculations.compute_heritability(
                self._penetrances, mafs, model_order=self._order
            ),
            h,  # It does not be casted to rational because the solver doesn't end
        )
        eq2 = sympy.Eq(self._max_penetrance(), sympy.Integer(1))
        return [eq1, eq2]  # System as a list with the equations

    def find_max_prevalence_table(
        self,
        mafs: List[float],
        h: float,
        solve_timeout: Union[int, bool] = True,
        check: bool = True,
    ) -> pytoxo.ptable.PTable:
        """Computes the table whose prevalence is maximum for the given MAFs
        and heritability, and returns it within a `PTable` object.

        Parameters
        ----------
        mafs : list[float]
            Array of floats representing the MAF of each locus.
        h : float
            Heritability of the table.
        solve_timeout : Union[int, bool], optional (default True)
            A maximum timeout, as seconds, for the solver to try to resolve the
            model. If it is exceeded, the operation will be aborted. Default
            (passed as 'True') is assumed heuristically. Pass as 'False' or
            'None' to do not use timeout.
        check: bool (default True)
            Flag to control when check the solution before return it. Default
            `True`.

        Returns
        -------
        pytoxo.ptable.PTable
            Penetrance table obtained within a `PTable` object.

        Raises
        ------
        pytoxo.errors.ResolutionError
            If PyToxo is not able to solve the model.
        pytoxo.errors.UnsolvableModelError
            If the model has not solution.
        ValueError
            With a bad parameter configuration.
        """
        self.check_find_table_parameters(mafs, h, solve_timeout, check)

        eq_system = self._build_max_prevalence_system(mafs, h)

        # Delegates to the solve method
        sol = self._solve(eq_system, solve_timeout)

        # Check the solution before return the tables
        if check:
            ok, delta = self._check_solution(eq_system, sol)
            if not ok:
                raise pytoxo.errors.ResolutionError(
                    f"The calculated solution is not correct. Difference of {delta}"
                )

        # Return the final achieved solution as penetrance table object
        return pytoxo.ptable.PTable(
            model_order=self._order,
            model_genotypes=self.calculate_genotypes(),
            model_penetrances=self._penetrances,
            values={
                self._variables[0]: sol[self._variables[0]],
                self._variables[1]: sol[self._variables[1]],
            },
            mafs=mafs,
        )

    def _build_max_heritability_system(
        self, mafs: List[float], p: float
    ) -> List[sympy.Eq]:
        """Builds the system of equations for this model, for a maximum
        heritability, for the given MAFs and prevalence.

        Parameters
        ----------
        mafs : list[float]
            Array of floats representing the MAF of each locus.
        p : float
            Prevalence of the table.

        Returns
        -------
        list[sympy.Eq]
            System of equations to solve this model for a maximum
            heritability, for the given MAFs and prevalence.
        """
        eq1 = sympy.Eq(
            pytoxo.calculations.compute_prevalence(
                self._penetrances, mafs, model_order=self._order
            ),
            p,  # It does not be casted to rational because the solver doesn't end
        )
        eq2 = sympy.Eq(self._max_penetrance(), sympy.Integer(1))
        return [eq1, eq2]  # System as a list with the equations

    def find_max_heritability_table(
        self,
        mafs: List[float],
        p: float,
        solve_timeout: Union[int, bool] = True,
        check: bool = True,
    ) -> pytoxo.ptable.PTable:
        """Computes the table whose heritability is maximum for the given MAFs
        and prevalence, and returns it within a `PTable` object.

        Parameters
        ----------
        mafs : list[float]
            Array of floats representing the MAF of each locus.
        p : float
            Prevalence of the table.
        solve_timeout : Union[int, bool], optional (default True)
            A maximum timeout, as seconds, for the solver to try to resolve the
            model. If it is exceeded, the operation will be aborted. Default
            (passed as 'True') is assumed heuristically. Pass as 'False' or
            'None' to do not use timeout.
        check: bool (default True)
            Flag to control when check the solution before return it. Default
            `True`.

        Returns
        -------
        pytoxo.ptable.PTable
            Penetrance table obtained within a `PTable` object.

        Raises
        ------
        pytoxo.errors.ResolutionError
            If PyToxo is not able to solve the model.
        pytoxo.errors.UnsolvableModelError
            If the model has not solution.
        ValueError
            With a bad parameter configuration.
        """
        self.check_find_table_parameters(mafs, p, solve_timeout, check)

        eq_system = self._build_max_heritability_system(mafs, p)

        # Delegates to the solve method
        sol = self._solve(eq_system, solve_timeout)

        # Check the solution before return the tables
        if check:
            ok, delta = self._check_solution(eq_system, sol)
            if not ok:
                raise pytoxo.errors.ResolutionError(
                    f"The calculated solution is not correct. Difference of {delta}"
                )

        # Return the final achieved solution as penetrance table object
        return pytoxo.ptable.PTable(
            model_order=self._order,
            model_genotypes=self.calculate_genotypes(),
            model_penetrances=self._penetrances,
            values={
                self._variables[0]: sol[self._variables[0]],
                self._variables[1]: sol[self._variables[1]],
            },
            mafs=mafs,
        )

    def check_find_table_parameters(
        self,
        mafs: List[float],
        h_or_p: float,
        solve_timeout: Union[int, bool] = True,
        check: bool = True,
    ) -> None:
        """Check the input parameters for a find table operation.

        Parameters
        ----------
        mafs : list[float]
            Array of floats representing the MAF of each locus.
        h_or_p : float
            Heritability or prevalence of the table.
        solve_timeout : Union[int, bool], optional (default True)
            A maximum timeout, as seconds, for the solver to try to resolve the
            model. If it is exceeded, the operation will be aborted. Default
            (passed as 'True') is assumed heuristically. Pass as 'False' or
            'None' to do not use timeout.
        check: bool (default True)
            Flag to control when check the solution before return it. Default
            `True`.

        Raises
        ------
        ValueError
            With a bad parameter configuration.
        """
        if type(mafs) is not list:
            raise ValueError("MAFs should be a list of floats.")
        for maf in mafs:
            if type(maf) is not float:
                raise ValueError("MAFs should be a list of floats.")
        if len(mafs) != self._order:
            raise ValueError("The MAFs have to be as many as the order has the model.")
        for maf in mafs:
            if not 0 <= maf <= 0.5:
                raise ValueError("MAFs should be a probability between 0 and 0.5.")

        if type(h_or_p) is not float:
            raise ValueError("Heritability or prevalence should be a float.")
        if not 0 <= h_or_p <= 1:
            raise ValueError(
                "Heritability or prevalence should be probability. I.e. a float between 0 and 1."
            )

        if type(solve_timeout) is not int and type(solve_timeout) is not bool:
            raise ValueError("Timeout parameter should be an integer or a boolean.")
        if type(solve_timeout) is int and solve_timeout < 0:
            raise ValueError("Timeout should be a positive integer.")

        if type(check) is not bool:
            raise ValueError("The check flag should be a boolean.")



# -*- coding: utf-8 -*-

###########################################################
# PyToxo
#
# A Python tool to calculate penetrance tables for 
# high-order epistasis models
#
# Copyright 2021 Borja González Seoane
#
# Contact: borja.gseoane@udc.es
###########################################################

"""Command line interface entry point."""

import argparse
import sys

import colorama
import termcolor

import pytoxo
import pytoxo.errors

# Necessary to use colors in Windows machines
colorama.init()

# Some predefined printing chunks
ok_hd = f"{termcolor.colored('[OK]', 'green', attrs=['bold'])}:"
warning_hd = f"{termcolor.colored('[WARNING]', 'yellow', attrs=['bold'])}:"
error_hd = f"{termcolor.colored('[ERROR]', 'red', attrs=['bold'])}:"


def main():
    # Define the parser
    parser = argparse.ArgumentParser(
        prog="pytoxo",
        description="PyToxo: A Python library for calculating penetrance "
        "tables of any bivariate epistasis model.",
        epilog="PyToxo CLI. Copyright 2021 Borja González Seoane",
    )
    parser.add_argument(
        "model",
        metavar="MODEL",
        type=str,
        nargs=1,
        help="The path to the epistatic model CSV file.",
    )
    parser.add_argument(
        "--gametes",
        action="store_true",
        help="This flag is to use GAMETES format to print the final table, "
        "instead of default CSV format.",
    )
    prev_or_her_flag = parser.add_mutually_exclusive_group(required=True)
    prev_or_her_flag.add_argument(
        "--max_prev",
        action="store_true",
        help="This flag set PyToxo to maximize the prevalence, so the "
        "'PREV_OR_HER' argument will be considered as heritability.",
    )
    prev_or_her_flag.add_argument(
        "--max_her",
        action="store_true",
        help="This flag set PyToxo to maximize the heritability, so the "
        "'PREV_OR_HER' argument will be considered as prevalence.",
    )
    parser.add_argument(
        "prev_or_her_arg",
        metavar="PREV_OR_HER",
        type=float,
        nargs=1,
        help="The heritability or prevalence to fix, depending of the used "
        "flag. Maximizing prevalence, this argument will be interpreted "
        "as heritability; and maximizing heritability, as prevalence.",
    )
    parser.add_argument(
        "mafs",
        metavar="MAFS",
        type=float,
        nargs="+",
        help="The MAFs to use. As many as the order has the model, followed.",
    )

    # Try to parse
    try:
        args = parser.parse_args()
    except argparse.ArgumentError:
        print(f"{error_hd} Argument parsing error.")
        sys.exit(1)

    # Get arguments
    model = args.model[0]
    mafs = args.mafs
    if args.max_prev:
        calc_target = pytoxo.Model.find_max_prevalence_table
    else:
        calc_target = pytoxo.Model.find_max_heritability_table
    prev_or_her = args.prev_or_her_arg[0]
    if args.gametes:
        output_format = "gametes"
    else:
        output_format = "csv"  # Default

    # Try to build model
    try:
        model = pytoxo.Model(filename=model)
    except IOError as e:
        print(f"{error_hd} {e.message}")
        sys.exit(1)
    except pytoxo.errors.BadFormedModelError as e:
        print(f"{error_hd} {e.message}")
        sys.exit(1)

    # Try to solve the model
    try:
        ptable = calc_target(model, mafs, prev_or_her)
    except pytoxo.errors.ResolutionError as e:
        print(f"{error_hd} {e.message}")
        sys.exit(1)
    except pytoxo.errors.UnsolvableModelError as e:
        print(f"{error_hd} {e.message}")
        sys.exit(1)
    except ValueError as e:
        print(f"{error_hd} Bad parameter configuration. {e.__str__()}")
        sys.exit(1)

    # Print the penetrance table to can integrate it in a typical Shell routine
    try:
        ptable.print_table(format=output_format)
    except pytoxo.errors.GenericCalculationError as e:  # Improvable exception
        print(f"{error_hd} {e.message}")
        sys.exit(1)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

###########################################################
# PyToxo
#
# A Python tool to calculate penetrance tables for 
# high-order epistasis models
#
# Copyright 2021 Borja González Seoane
#
# Contact: borja.gseoane@udc.es
###########################################################

"""PyToxo error definitions."""

from typing import Dict, Tuple, Union


class BadFormedModelError(Exception):
    """Representation of a parse error with an epistatic model."""

    def __init__(
        self,
        error_object: Union[str, Dict[str, str], Tuple[list, list]],
        cause: str = "Bad formed model",
    ):
        """Creates an exception of a parse error with an epistatic model.

        Parameters
        ----------
        error_object : Union[str, dict[str, str], (list, list)]
            The input model which causes the error in one of its possible forms.
        cause : str, optional
            The error cause during parsing process (default "Bad formed file").
        """
        self.error_object = error_object
        self.cause = cause
        self.message = f"Bad formed model with '{error_object}' due to: '{cause}'."


class ResolutionError(Exception):
    """Representation of a resolution error of PyToxo."""

    def __init__(
        self,
        cause: str = "PyToxo can not solve this model",
        model_name: str = None,
        equation: str = None,
        message: str = None,
    ):
        """Creates an exception of a solving error with a PyToxo model.

        Parameters
        ----------
        cause : str, optional
            The cause during solve process (default "PyToxo can not solve
            this model").
        model_name : str, optional
            The model given name.
        equation : str, optional
            The equation that cause the resolution error, as string. Default is
            none.
        message : str, optional
            The message to print, directly passed here. Other arguments would be
            ignored if this one is used.
        """
        msg_header = "Resolution error: "
        self.model_name = model_name
        self.cause = cause
        self.equation = equation
        self.message = message

        if not self.message:
            if not self.model_name:
                model_msg_text = "Model"
            else:
                model_msg_text = f"Model '{self.model_name}'"
            if self.equation:
                self.message = (
                    f"{msg_header} {model_msg_text} due to '{cause}'. "
                    f"Problematic equation: '{self.equation}'."
                )
            else:
                self.message = f"{msg_header} '{cause}'."
        else:
            self.message = f"{msg_header} {self.message}."


class UnsolvableModelError(Exception):
    """Representation of an unsolvable model error of PyToxo."""

    def __init__(
        self, model_name: str = None, equation: str = None, message: str = None
    ):
        """Creates an exception of an unsolvable model.

        Parameters
        ----------
        model_name : str, optional
            The model given name.
        equation : str, optional
            The equation that cause the resolution error, as string. Default is
            none.
        message : str, optional
            The message to print, directly passed here. Other arguments would be
            ignored if this one is used.
        """
        msg_header = "Unsolvable model:"
        self.model_name = model_name
        self.equation = equation
        self.message = message

        if not self.message:
            if self.model_name:
                model_msg_text = "model"
            else:
                model_msg_text = f"model ('{self.model_name}')"
            if self.equation:
                self.message = (
                    f"{msg_header} This {model_msg_text} has not solution. "
                    f"Equation: '{self.equation}'."
                )
            else:
                self.message = f"{msg_header} This {model_msg_text} has not solution."
        else:
            self.message = f"{msg_header} {self.message}."


class GenericCalculationError(Exception):
    """Representation of a generic calculation error of PyToxo."""

    def __init__(self, function: str = None):
        """Creates a generic calculation error of PyToxo.

        Parameters
        ----------
        function : str, optional
            The name of the function when the error occurs.
        """
        msg_skeleton = "Calculation error produced in: '{}'"
        self.function = function

        if not self.function:
            self.message = "Calculation error produced"
        else:
            self.message = msg_skeleton.format(self.function)


class GUIUnsupportedPlatformError(Exception):
    """Representation of a GUI platform support error of PyToxo."""

    def __init__(self, platform: str):
        """Creates a GUI platform support error of PyToxo.

        Parameters
        ----------
        platform : str
            The name of the platform where the GUI is used: Darwin, Linux,
            Windows or other unsupported one.
        """
        if platform == "Linux":
            self.message = (
                f"Your plaform (detected {platform}) supports PyToxo's GUI, "
                f"but something is wrong... Please, check you have Tk and "
                f"its Python binders installed and correctly reachable. "
                f"You can usually install them with the command:"
                f"\n\n\tsudo apt install python3-tk"
            )
        elif platform == "Darwin" or platform == "Windows":
            self.message = (
                f"Your plaform (detected {platform}) supports PyToxo's GUI, "
                f"but something is wrong... Please, check you have Tk and "
                f"its Python binders installed and correctly reachable."
            )
        else:
            self.message = (
                f"Your platform (detected {platform}) does not support the "
                f"PyToxo's GUI. Please, try the CLI instead."
            )



# -*- coding: utf-8 -*-

###########################################################
# PyToxo
#
# A Python tool to calculate penetrance tables for 
# high-order epistasis models
#
# Copyright 2021 Borja González Seoane
#
# Contact: borja.gseoane@udc.es
###########################################################

"""PyToxo graphical user interface package."""



# -*- coding: utf-8 -*-

###########################################################
# PyToxo
#
# A Python tool to calculate penetrance tables for 
# high-order epistasis models
#
# Copyright 2021 Borja González Seoane
#
# Contact: borja.gseoane@udc.es
###########################################################

from pytoxo.model import Model
from pytoxo.ptable import PTable

# Package PyToxo exports only the `Model` class
__all__ = ["Model", "PTable"]



# -*- coding: utf-8 -*-

###########################################################
# PyToxo
#
# A Python tool to calculate penetrance tables for 
# high-order epistasis models
#
# Copyright 2021 Borja González Seoane
#
# Contact: borja.gseoane@udc.es
###########################################################

"""Penetrance table definition."""

import os

import numpy
import sympy
from typing import List, Dict

import pytoxo.calculations
import pytoxo.errors


class PTable:
    """Representation of a penetrance table."""

    def __init__(
        self,
        model_order: int,
        model_genotypes: List[str],
        model_penetrances: List[sympy.Expr],
        values: Dict[sympy.Symbol, float],
        mafs: List[float] = None,
        model_name: str = None,
    ):
        """Creates a penetrance table from a given PyToxo model defined by
        its variables and penetrances, and its variable values.

        Parameters
        ----------
        model_order : int
            Order from the PyToxo model from which to create the penetrance
            table.
        model_genotypes : List[str]
            List of the genotype definitions like generated by
            `_calculate_genotypes` PyToxo model method.
        model_penetrances : List[sympy.Expr]
            Penetrances from the PyToxo model from which to create the
            penetrance table.
        values : dict[sympy.Symbol, float]
             Value for each of the variables represented in model, typically
             `x` and `y`. Keys are the own variables as `sympy.Symbol`.
        mafs : List[float], optional
             The MAFs used in the generation of the table. Optional
             parameter, only necessary to save then the table using GAMETES
             format.
        model_name: str, optional
            Optional value of the model which compound this table,
            to identify it easily.
        """
        self._order = model_order
        self._genotypes = model_genotypes
        self._penetrance_values = [
            p.subs(values) for p in model_penetrances
        ]  # Try to substitute `y` in expression `x` does not cause errors, simply are ignored
        self._values = list(values.values())  # Only used to save to GAMETES
        self._model_name = model_name
        self._mafs = mafs  # Only used to save to GAMETES
        self._prevalence = None  # Only used to save to GAMETES
        self._heritability = None  # Only used to save to GAMETES

    ########################################
    # Getters and setters for properties

    def _get_model_name(self) -> str:
        return self._model_name

    def _set_model_name(self, model_name: str) -> None:
        self._model_name = model_name

    model_name = property(_get_model_name, _set_model_name)

    def _get_mafs(self) -> List[float]:
        return self._mafs

    def _set_mafs(self, mafs: List[float]) -> None:
        self._mafs = mafs

    mafs = property(_get_mafs, _set_mafs)

    def _get_prevalence(self) -> float:
        return self._prevalence

    def _set_prevalence(self, prevalence: float) -> None:
        self._prevalence = prevalence

    prevalence = property(_get_prevalence, _set_prevalence)

    def _get_heritability(self) -> float:
        return self._heritability

    def _set_heritability(self, heritability: float) -> None:
        self._heritability = heritability

    heritability = property(_get_heritability, _set_heritability)

    @property
    def order(self) -> int:
        return self._order

    @property
    def genotypes(self) -> List[str]:
        return self._genotypes

    @property
    def genotypes_as_numpy(self) -> numpy.array:
        return numpy.array(self._genotypes)

    @property
    def penetrance_values(self) -> List[float]:
        return self._penetrance_values

    @property
    def penetrance_values_as_numpy(self) -> numpy.array:
        return numpy.array(self._penetrance_values)

    ########################################

    def __hash__(self):
        return hash(
            hash(self._order)
            + hash(str(self._genotypes))
            + hash(str(self._penetrance_values))
            + hash(self._values)
            + hash(self._model_name)
            + hash(self._mafs)
            + hash(self._prevalence)
            + hash(self._heritability)
        )

    def __eq__(self, other):
        return hash(self) == hash(other)

    def _compound_table_as_text(self) -> str:
        """Compound the penetrance table as text, to print it ot save into a
        file it.

        Returns
        -------
        str
            The penetrance table formatted as a text string.
        """
        # Generate lines of the file with genotypes and its penetrances
        lines = ""
        for genotype, penetrance in zip(self._genotypes, self._penetrance_values):
            lines += f"{''.join(genotype)},{penetrance}\n"

        return lines

    def _compound_table_as_gametes(self) -> str:
        """Compound the penetrance table as GAMETES format, to save it into a
        file.

        Attributes `mafs` must be filled.

        Returns
        -------
        str
            The penetrance table formatted as a text string.

        Raises
        ------
        GenericCalculationError
            Failing to calculate prevalence or heritability to save as GAMETES.
        """
        """Calculate prevalence and heritability. Do here to optimize,
            because they are only needed to save as GAMETES format"""
        try:
            self._prevalence = pytoxo.calculations.compute_prevalence(
                penetrances=self._penetrance_values,
                mafs=self._mafs,
                model_order=self._order,
            )
            self._heritability = pytoxo.calculations.compute_heritability(
                penetrances=self._penetrance_values,
                mafs=self._mafs,
                model_order=self._order,
            )
        except pytoxo.errors.GenericCalculationError as e:
            raise pytoxo.errors.GenericCalculationError(
                "It is no possible to calculate the "
                f"prevalence or the heritability using {e.function}."
                "Are you using the original used MAFs?"
            )

        gametes_skeleton = (
            "Attribute names:\t{}\nMinor allele "
            "frequencies:\t{}\nx: {}\ny: {}\nPrevalence: {"
            "}\nHeritability: {}\n\nTable:\n\n{} "
        )

        # Prepare fields to fill
        attribute_names = "\t".join([f"P{i}" for i in range(0, self.order)])
        mafs = "\t".join([f"{'{:.3f}'.format(maf)}" for maf in self._mafs])
        x = str(self._values[0])
        y = str(self._values[1])
        prev = str(self._prevalence.evalf())
        her = str(self._heritability.evalf())

        # Prepare table to fill
        table = ""
        for i in range(0, len(self._penetrance_values), 9):
            table += (
                f"{self._penetrance_values[i]}, "
                f"{self._penetrance_values[i+1]}, "
                f"{self._penetrance_values[i+2]}\n"
                f"{self._penetrance_values[i+3]}, "
                f"{self._penetrance_values[i+4]}, "
                f"{self._penetrance_values[i+5]}\n"
                f"{self._penetrance_values[i+6]}, "
                f"{self._penetrance_values[i+7]}, "
                f"{self._penetrance_values[i+8]}\n\n"
            )
        table = table[:-1]

        # Fill skeleton and return
        return gametes_skeleton.format(attribute_names, mafs, x, y, prev, her, table)

    def print_table(self, format: str = "csv") -> None:
        """Print the penetrance as raw text.

        Currently formats CSV and GAMETES are available. To use GAMETES,
        this object has to be filled the attribute `mafs`.

        Parameters
        ----------
        format : str (default "csv")
            The format of the final print. Currently CSV format (`csv`
            flag) and GAMETES (`gametes` flag) are supported. Default is
            CSV.

        Raises
        ------
        ValueError
            If unsupported format tentative.
        GenericCalculationError
            Failing to calculate prevalence or heritability to print as GAMETES.
        """
        supported_formats = ["csv", "gametes"]

        # Input handling and checks
        format = format.lower()  # Defer
        if format not in supported_formats:
            raise ValueError(f"Unsupported '{format}' format")

        # Check if is possible to use GAMETES
        if format == "gametes" and self._mafs is None:
            raise ValueError(
                f"The '{format}' format requires to be filled the attribute `mafs`."
            )

        # Generate the table
        if format == "csv":
            table = self._compound_table_as_text()
        else:
            table = self._compound_table_as_gametes()

        print(table)

    def write_to_file(
        self, filename: str, overwrite: bool = False, format: str = "gametes"
    ) -> None:
        """Write the penetrance table into a file.

        Currently formats CSV and GAMETES are available. To use GAMETES,
        this object has to be filled the attribute `mafs`.

        Parameters
        ----------
        filename : str
            The full file name where writes the table.
        overwrite : bool (default False)
            A flag that should be passed as true to overwrite the final file
            if it already exists.
        format : str (default "gametes")
            The format of the final file. Currently CSV format (`csv`
            flag) and GAMETES (`gametes` flag) are supported. Default is
            GAMETES.

        Raises
        ------
        ValueError
            If unsupported format tentative.
        FileExistsError
            If the `filename` file already exists and the `overwrite` has not be
            passed as true.
        IsADirectoryError
            If `filename` is a existent directory.
        GenericCalculationError
            Failing to calculate prevalence or heritability to save as GAMETES.
        """
        supported_formats = ["gametes", "csv"]

        # Input handling and checks
        format = format.lower()  # Defer
        if format not in supported_formats:
            raise ValueError(f"Unsupported '{format}' format")

        # Check if is possible to use GAMETES
        if format == "gametes" and self._mafs is None:
            raise ValueError(
                f"The '{format}' format requires to be filled the attribute `mafs`."
            )

        # Calculate final filename
        filename = os.path.normpath(filename)
        # Check final file name
        if os.path.exists(filename) and not overwrite:
            raise FileExistsError(filename)
        if os.path.exists(filename) and not os.path.isfile(filename):
            raise IsADirectoryError(filename)

        # Generate the table
        if format == "gametes":
            table = self._compound_table_as_gametes()
        else:
            table = self._compound_table_as_text()

        # Write file
        if overwrite and os.path.exists(filename):
            os.remove(filename)
        with open(filename, "x") as f:
            f.write(table)



# -*- coding: utf-8 -*-

###########################################################
# PyToxo
#
# A Python tool to calculate penetrance tables for 
# high-order epistasis models
#
# Copyright 2021 Borja González Seoane
#
# Contact: borja.gseoane@udc.es
###########################################################

"""PyToxo util module."""

import functools
from threading import Thread


def timeout(timeout):
    """Timeout wrapper to associate a given timeout with a function. If the
    time is exceeded, a `TimeoutError` is raised. This approach works both in
    Unix-like and Windows machines. These last are more complicated because
    do not support signals well, which are the normally workaround to
    achieve these stuff.

    Raises
    ------
    TimeoutError
        If the configured timeout is exceeded.
    """

    def deco(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            res = [TimeoutError()]

            def target_helper():
                """Overrides the `res` variable, if `func` works."""
                res[0] = func(*args, **kwargs)

            t = Thread(target=target_helper)
            t.daemon = True
            try:
                t.start()
                t.join(timeout)
            except:
                """Possible but very unusual error handling multiprocessing.
                Ignore them seems enough workaround"""
                pass

            """Check if `res` has been overridden by `func`. If not, it signs 
            that `func` has been interrupted, so raise the time exceeded 
            error"""
            if isinstance(res[0], BaseException):
                raise TimeoutError
            return res[0]

        return wrapper

    return deco



# -*- coding: utf-8 -*-

###########################################################
# PyToxo
#
# A Python tool to calculate penetrance tables for 
# high-order epistasis models
#
# Copyright 2021 Borja González Seoane
#
# Contact: borja.gseoane@udc.es
###########################################################

"""PyToxo calculations module: some math-based methods used in different
parts of the library.

If possible, internal Sympy classes and methods are chosen instead of Python
built-ins for optimization.
"""

import functools
import inspect
import itertools
from typing import List, Union

from sympy import Integer, Rational, Expr, Add, Mul, Pow, nsimplify, simplify

import pytoxo.util
from pytoxo.errors import GenericCalculationError


def genotype_probabilities(
    mafs: Union[List[Rational], List[float]], model_order: int = None
) -> Union[List[Rational], List[Expr]]:
    """Computes the probabilities associated with all genotype combinations
    given each MAF (minor allele frequency).

    The genotype frequency can be calculated from the allele frequencies. It
    is usually assumed for its calculation that the Hardy-Weinberg equilibrium
    is fulfilled, that is, if the frequency of allele `A` is `M` and the
    frequency of allele `a` is `m`, the probability of having the `AA` genotype
    is `M^2`, having `Aa` is `2 * M * m` and having `aa` is `m^2` (this could be
    extended to any order of interaction, including the probabilities of each
    allele in the products).

    Therefore, given the MAF (minor allele frequency), the frequencies of all
    possible genotypes can be calculated. For example, if MAF is 0.25, then
    `M = 0.25`, and `m = 1 - 0.25 = 0.75`. From there, the probabilities of
    all possible allele combinations (of all possible genotypes) can be
    calculated.

    Parameters
    ----------
    mafs : Union[list[Rational], list[float]]
        Minor allele frequencies array. Accept rationals and floats.
    model_order : int, optional
        The order of the implicated model. This optional parameter is used to
        optimize the computation. Default is `None` and supposes a generalist
        optimization strategy, less optimal.

    Returns
    -------
    Union[list[Rational], list[Expr]]
        Array with the probabilities of all possible allele combinations as
        rational numbers.

    Raises
    ------
    GenericCalculationError
        On any error situation.
    """
    try:
        mafs = [nsimplify(m) for m in mafs]  # Assert rationals

        af_zip = []  # Zip with pairs of alleles frequencies as `(m, M)` tuples
        for maf in mafs:
            m = maf
            M = Add(Integer(1), Mul(Integer(-1), maf))
            af_zip.append((m, M))

        gen_probs = (
            []
        )  # Genotype probabilities as `[M ** 2, 2 * M * m, m ** 2]` sub-lists
        for allele_pair in af_zip:
            m = allele_pair[0]
            M = allele_pair[1]
            gen_probs.append(
                [Pow(M, Integer(2)), Mul(Integer(2), M, m), Pow(m, Integer(2))]
            )

        # Cartesian product of the probabilities
        gen_probs_cps = list(itertools.product(*gen_probs))

        # Reduction of the cartesian product with a multiplication for each sub-list
        gen_probs_cp_red = []
        for gen_probs_cp in gen_probs_cps:
            gen_probs_cp_red.append(functools.reduce(Mul, gen_probs_cp, Integer(1)))

        return gen_probs_cp_red

    except:
        frame = inspect.currentframe()
        raise GenericCalculationError(inspect.getframeinfo(frame).function)


def compute_prevalence(
    penetrances: Union[List[Expr], List[Rational], List[float]],
    mafs: Union[List[Rational], List[float]] = None,
    gp: Union[List[Rational], List[float]] = None,
    model_order: int = None,
) -> Union[Rational, Expr]:
    """Tries to compute the prevalence for a given penetrance list.

    One of `mafs` and `gp` is required. Is the both are used, `mafs` is ignored.

    Parameters
    ----------
    penetrances : Union[list[Expr], list[Rational], list[float]]
        Penetrance values array.
    mafs : Union[list[Rational], list[float]], optional
        Minor allele frequencies array. Ignored if `gp` is used.
    gp : Union[list[Rational], list[float]], optional
        Genotype probabilities array.
    model_order : int, optional
        The order of the implicated model. This optional parameter is used to
        optimize the computation. Default is `None` and supposes a generalist
        optimization strategy, less optimal.

    Returns
    -------
    Union[Rational, Expr]
        Prevalence of the penetrance table. Returns a rational if it is
        possible to solve the expression numerically and the expression if not.

    Raises
    ------
    ValueError
        If none of `mafs` and `gp` are passed as parameter.

    Raises
    ------
    GenericCalculationError
        On any error situation.
    """
    try:
        if not gp:
            if not mafs:
                raise ValueError(
                    "One of `mafs` and `gp` is required. Is the both are used, `mafs` is ignored."
                )
            else:
                gp = genotype_probabilities(mafs)
        else:
            gp = [nsimplify(p) for p in gp]  # Assert rationals

        penetrances = [nsimplify(p) for p in penetrances]  # Assert rationals

        # `penetrances .* gp`
        prods = []
        for pen, prob in zip(penetrances, gp):
            prods.append(Mul(pen, prob))

        addition = Add(*prods)  # Addition of all the elements of the product array

        """Try to solve. Innocuous if the expression is still symbolic and saves a
        lot of time if we are dealing with an expression that is already reducible 
        to a numeric value"""
        addition = addition.evalf().nsimplify()

        return _try_to_simplify(
            addition, model_order
        )  # Return simplified to optimize next steps

    except:
        frame = inspect.currentframe()
        raise GenericCalculationError(inspect.getframeinfo(frame).function)


def compute_heritability(
    penetrances: Union[List[Expr], List[Rational], List[float]],
    mafs: Union[List[Rational], List[float]],
    model_order: int = None,
) -> Union[Rational, Expr]:
    """Tries to compute the heritability for a given penetrance table
    defined by its values.

    Parameters
    ----------
    penetrances : Union[list[Expr], list[Rational], list[float]]
        Penetrance values array.
    mafs : Union[list[Rational], list[float]], optional
        Minor allele frequencies array.
    model_order : int, optional
        The order of the implicated model. This optional parameter is used to
        optimize the computation. Default is `None` and supposes a generalist
        optimization strategy, less optimal.

    Returns
    -------
    Union[Rational, Expr]
        Heritability of the penetrance table. Returns a rational if it is
        possible to solve the expression numerically and the expression if not.

    Raises
    ------
    GenericCalculationError
        On any error situation.
    """
    try:
        penetrances = [nsimplify(p) for p in penetrances]  # Assert rationals

        gp = genotype_probabilities(mafs)
        p = compute_prevalence(penetrances, mafs, gp)

        # `(penetrances - p).^2 .* gp`
        prods = []
        for pen, prob in zip(penetrances, gp):
            prods.append(Mul(Pow(Add(pen, Mul(Integer(-1), p)), Integer(2)), prob))

        # Denominator of the final expression
        denom = _try_to_simplify(
            Pow(Mul(p, Add(Integer(1), Mul(Integer(-1), p))), Integer(-1), model_order)
        )

        # `prods / denom`, because `denom` is a negative pow
        mult = Mul(Add(*prods), denom)

        """Try to solve. Innocuous if the expression is still symbolic and saves a
        lot of time if we are dealing with an expression that is already reducible 
        to a numeric value"""
        mult = mult.evalf().nsimplify()

        return _try_to_simplify(
            mult, model_order
        )  # Return simplified to optimize next steps

    except:
        frame = inspect.currentframe()
        raise GenericCalculationError(inspect.getframeinfo(frame).function)


def _try_to_simplify(expr: Expr, timeout: int = 60, model_order: int = None):
    """Help function which encapsulate the a call to an attempt to simplify an
    expression with a maximum timeout. If the expression could not be simplified
    in that time, it returns it as it was at the beginning, else return it
    simplified. In this way it is ensured that the simplifications of this
    module help the efficiency of the program and not the opposite. It also
    prevents the program from seemingly stuck trying to simplify overly
    complex expressions.

    Parameters
    ----------
    expr : Expr
        Expression to simplify.
    timeout : int, optional (default 60)
        Maximum timeout for the try, as seconds. Default 1 minute as 60 seconds.
        If the `model_order` parameter is used, a better value for this is
        also achieved. Use `None` to don't use a timeout.
    model_order : int, optional
        The order of the implicated model. This optional parameter is used to
        optimize the computation. Default is `None` and supposes a generalist
        optimization strategy, less optimal.

    Returns
    -------
    Expr
        The input expression, simplified or not.
    """
    if model_order:
        timeout = model_order * 60  # Seconds

    @pytoxo.util.timeout(timeout)
    def simplify_call():
        return simplify(expr)

    try:
        return simplify_call()
    except TimeoutError:
        return expr



# -*- coding: utf-8 -*-

###########################################################
# PyToxo
#
# A Python tool to calculate penetrance tables for
# high-order epistasis models
#
# Copyright 2021 Borja González Seoane
#
# Contact: borja.gseoane@udc.es
###########################################################

import setuptools


with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

with open("requirements_with_gui.txt") as f:
    requirements = f.read().splitlines()


version = "1.1"

setuptools.setup(
    name="pytoxo",
    version=version,
    packages=["pytoxo", "pytoxo_cli", "pytoxo_gui"],
    entry_points={
        "console_scripts": [
            "pytoxo_cli = pytoxo_cli.__main__:main",
            "pytoxo_gui = pytoxo_gui.__main__:main",
        ],
    },
    python_requires=">=3.8",
    install_requires=requirements,
    data_files=[
        ("", ["requirements_with_gui.txt"]),
        ("", ["README.md"]),
        ("", ["LICENSE"]),
    ],
    url="https://github.com/bglezseoane/pytoxo",
    download_url=f"https://github.com/bglezseoane/pytoxo/archive/{version}.tar.gz",
    license="LICENSE",
    author="Borja González Seoane",
    author_email="borja.gseoane@udc.es",
    description="A Python tool to calculate penetrance tables for high-order epistasis models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Operating System :: OS Independent",
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
)
